
# Copyright (c) 2023 Watsen Networks.  All Rights Reserved.

_A='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2022-05-26",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2022-05-26",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-1",\n        "revision": "2022-05-26",\n        "feature": [\n        ],\n        "namespace": "https://watsen.net/sztpd-1",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
import json
from .yangcore.yl import yangcore_library
from .yangcore.utils import set_yyyy_mm_dd
def nbi_1():A=yangcore_library+common_nbi_snippet+bss_1_vZ;B=json.loads(A);B=set_yyyy_mm_dd(B);A=json.dumps(B);return A
def nbi_x():A=yangcore_library+common_nbi_snippet+bss_x_vZ;B=json.loads(A);B=set_yyyy_mm_dd(B);A=json.dumps(B);return A
def nbi_x_tenant():A=yangcore_library+common_nbi_snippet+bss_x_tenant_vZ;B=json.loads(A);B=set_yyyy_mm_dd(B);A=json.dumps(B);return A
def sbi_rfc8572():return sbi_rfc8572_yl_str
def sbi_legacy():return'FIXME'
sbi_rfc8572_yl_str='\n{\n  "ietf-yang-library:modules-state": {\n    "module-set-id": "TBD",\n    "module": [\n      {\n        "name": "ietf-yang-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-inet-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-inet-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-netconf-acm",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-netconf-acm",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-datastores",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-datastores",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-yang-library",\n        "revision": "2019-01-04",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-library",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-sztp-bootstrap-server",\n        "revision": "2019-04-30",\n        "feature": [\n          "onboarding-server"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-sztp-bootstrap-server",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-yang-structure-ext",\n        "revision": "2019-03-07",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-structure-ext",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-crypto-types",\n        "revision": "2022-05-26",\n        "feature": [\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-crypto-types",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-ztp-types",\n        "revision": "2022-05-26",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-ztp-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-sztp-csr",\n        "revision": "2022-05-26",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-sztp-csr",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}'
common_nbi_snippet='\n      {\n        "name": "ietf-sztp-conveyed-info",\n        "revision": "2019-04-30",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-sztp-conveyed-info",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-x509-c2n",\n        "revision": "YYYY-MM-DD",\n        "namespace": "https://watsen.net/wnc2n",\n        "conformance-type": "implement"\n      },\n'
bss_1_vZ=_A
bss_x_vZ='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2022-05-26",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2022-05-26",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-1",\n        "revision": "2022-05-26",\n        "feature": [\n        ],\n        "namespace": "https://watsen.net/sztpd-1",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-sztpd-x",\n        "revision": "2022-05-26",\n        "feature": [\n        ],\n        "namespace": "https://watsen.net/sztpd-x",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
bss_x_tenant_v1='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2022-05-26",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2022-05-26",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-1",\n        "revision": "2022-05-26",\n        "feature": [\n          "system-level-administration-disabled"\n        ],\n        "namespace": "https://watsen.net/sztpd-1",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
bss_x_tenant_vZ=_A